#include "yans-wifi-phy-ap.h"
#include "yans-wifi-channel.h"
#include "wifi-mode.h"
#include "wifi-preamble.h"
#include "wifi-phy-state-helper.h"
#include "error-rate-model.h"
#include "ns3/simulator.h"
#include "ns3/packet.h"
#include "ns3/assert.h"
#include "ns3/log.h"
#include "ns3/double.h"
#include "ns3/uinteger.h"
#include "ns3/enum.h"
#include "ns3/pointer.h"
#include "ns3/net-device.h"
#include "wifi-mac.h"
#include "ap-wifi-mac.h"
#include "sta-wifi-mac.h"
#include "adhoc-wifi-mac.h"
#include "wifi-net-device.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/boolean.h"
#include "snr-tag.h"
#include "signal-tag.h"
#include "rate-tag.h"
#include "noise-tag.h"
#include "yans-wifi-phy.h"
#include <cmath>
#include "ns3/ipv4-header.h" 
#include "ns3/core-module.h"
#include <string.h>

#define VARIATION_THRES 1
#define HIGH 1
#define LOW 0
#define PACKET_LOST 2

NS_LOG_COMPONENT_DEFINE ("YansWifiPhyAp");

namespace ns3 {

    
NS_OBJECT_ENSURE_REGISTERED (YansWifiPhyAp);

bool fromSta(const Ptr<const Packet> &packet);

TypeId
YansWifiPhyAp::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::YansWifiPhyAp")
    .SetParent<YansWifiPhy> ()
    .AddConstructor<YansWifiPhyAp> ()
    .AddTraceSource("MechanismEvent",
                    "Trace Source used for the mechanism",
                     MakeTraceSourceAccessor(&YansWifiPhyAp::m_phyMechanismTrace)) 
    .AddTraceSource("RateMechanismEvent",
                    "Trace Source used for the rate mechanism",
                     MakeTraceSourceAccessor(&YansWifiPhyAp::m_phyRateMechanismTrace)) 
  ;
  return tid;
}

bool YansWifiPhyAp::isEnable()
{
    return m_enable;
}

void YansWifiPhyAp::setEnable(bool b)
{
    m_enable=b;
}



YansWifiPhyAp::YansWifiPhyAp () :YansWifiPhy()
{
    NS_LOG_FUNCTION (this);
    m_drop = 0;
    m_lastSnr = 0;
    m_signal = 0;
    m_packetsReceived=0;
    m_enable=true;
    m_packetCounter=0;
  
    m_received_corr=0;
    m_sent=0;
        /*Initialize Mechanism Rates*/
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate6_5MbpsBW20MHz(), 5.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate13MbpsBW20MHz(), 8.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate19_5MbpsBW20MHz(), 10.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate26MbpsBW20MHz(), 14.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate39MbpsBW20MHz(), 17.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate52MbpsBW20MHz(), 22.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate58_5MbpsBW20MHz(), 23.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate65MbpsBW20MHz(), 24.0));
    
    currentRate = WifiPhy::GetOfdmRate65MbpsBW20MHz();
    
    SetTxPowerStart(20);
    SetTxPowerEnd(20);    
}

YansWifiPhyAp::~YansWifiPhyAp ()
{
    NS_LOG_DEBUG("AP RECEIVED:"<<m_received_corr);
    NS_LOG_DEBUG("AP SENT:"<<m_sent);

  NS_LOG_FUNCTION (this);
}

InterferenceHelper::SnrPer
YansWifiPhyAp::GetSnrPer (Ptr<InterferenceHelper::Event> event)
{
    return m_interference.CalculateSnrPer(event);
}

void
YansWifiPhyAp::SendPacket (Ptr<const Packet> packet, WifiTxVector txVector, enum WifiPreamble preamble)
{
    if(m_lastSnr!=0 )
    {
        SnrTag snrTag;            
        snrTag.Set( m_lastSnr );
        if (! packet->PeekPacketTag (snrTag))
            packet->AddByteTag (snrTag);            
                
        SignalTag signalTag;
        signalTag.Set(m_signal);
        if (! packet->PeekPacketTag (signalTag))
                packet->AddByteTag (signalTag);
    } 
    m_sent++;
    YansWifiPhy::SendPacket(packet,txVector,preamble);
}

double YansWifiPhyAp::getLastSnr()
{
    return m_lastSnr;
}

void
YansWifiPhyAp::NotifyRateMechanism(Ptr<const Packet> packet, Ptr<InterferenceHelper::Event> event, int action){
      m_phyRateMechanismTrace(packet, this, event, action);
}

void
YansWifiPhyAp::EndReceive (Ptr<Packet> packet, Ptr<InterferenceHelper::Event> event)
{
  NS_LOG_FUNCTION (this << packet << event);
  NS_ASSERT (IsStateRx ());
  NS_ASSERT (event->GetEndTime () == Simulator::Now ());

  struct InterferenceHelper::SnrPer snrPer;
  snrPer = m_interference.CalculateSnrPer (event);  
  m_interference.NotifyRxEnd ();
  bool packetfromSta=fromSta(packet);
  Ptr<Packet> p=packet->Copy();
    bool fail=false;

//   NS_LOG_DEBUG ("mode=" << (event->GetPayloadMode ().GetDataRate ()) <<
//                 ", snr=" << snrPer.snr << ", per=" << snrPer.per << ", size=" << packet->GetSize ());
  if (m_random->GetValue () > snrPer.per)
    {
      NotifyRxEnd (packet);
      uint32_t dataRate500KbpsUnits = event->GetPayloadMode ().GetDataRate () * event->GetTxVector().GetNss()/ 500000;
      bool isShortPreamble = (WIFI_PREAMBLE_SHORT == event->GetPreambleType ());
      double signalDbm = RatioToDb (event->GetRxPowerW ()) + 30;
      double noiseDbm = RatioToDb (event->GetRxPowerW () / snrPer.snr) - GetRxNoiseFigure () + 30;
      NotifyMonitorSniffRx (packet, (uint16_t)GetChannelFrequencyMhz (), GetChannelNumber (), dataRate500KbpsUnits, isShortPreamble, signalDbm, noiseDbm);
      m_state->SwitchFromRxEndOk (packet, snrPer.snr, event->GetPayloadMode (), event->GetPreambleType ());
            
      if(packetfromSta)
      {
         m_lastSnr=RatioToDb( snrPer.snr);
        m_packetsReceived++;
        m_signal = signalDbm;
      }
    }
  else
    {
      /* failure. */
      fail=true;
      NotifyRxDrop (packet);
      m_state->SwitchFromRxEndError (packet, snrPer.snr);
      
      if(packetfromSta)
      {
        m_drop++;
        NotifyRateMechanism(packet,event, PACKET_LOST);
      }
    }
    
//     
    if(packetfromSta)
    {
        m_packetCounter++;
        if(!fail)
            m_received_corr++;
        
        checkDeviation(p,event);
    }    
}

void YansWifiPhyAp::checkDeviation(Ptr<Packet> &packet,Ptr<InterferenceHelper::Event> &event)
{    
    SignalTag snrVariationTag;    
    
    if(m_enable)
    {
        if(packet->FindFirstMatchingByteTag(snrVariationTag) )
        {
            int r=(int)snrVariationTag.Get();
            setRate( r);
        }
        else if(m_packetsReceived > 1000 )
        {
            increaseRate();
            m_packetsReceived = 0;
        }
    }        
        NotifyRateMechanism(packet, event, LOW);//LOW has no meaning
    
}

double YansWifiPhyAp::getVariation()
{
    return m_variation;
}

void YansWifiPhyAp::increaseRate()
{
    MechanismRate newRate =rates.at(rates.size()-1);
    for (uint32_t i = 0; i < rates.size()-1; i++) 
    {
        if (rates.at(i).GetRate() == currentRate )
        {
            newRate=rates.at(i + 1);
        }
    }
    PointerValue tmp;
    m_device->GetAttribute("RemoteStationManager",tmp);
    Ptr<Object> remStationM = tmp.GetObject ();
    remStationM->SetAttribute("DataMode",StringValue(newRate.GetRate().GetUniqueName()) );
    remStationM->SetAttribute("ControlMode",StringValue(newRate.GetRate().GetUniqueName()) );
    currentRate=newRate.GetRate();
}

void YansWifiPhyAp::setRate(int r)
{
    MechanismRate rate=rates.at(r);
    PointerValue tmp;
    m_device->GetAttribute("RemoteStationManager",tmp);
    Ptr<Object> remStationM = tmp.GetObject ();
    remStationM->SetAttribute("DataMode",StringValue(rate.GetRate().GetUniqueName()) );
    remStationM->SetAttribute("ControlMode",StringValue(rate.GetRate().GetUniqueName()) );
    currentRate=rate.GetRate();

}


bool fromSta(const Ptr<const Packet> &packet) {
    
    WifiMacHeader h;
    Mac48Address addr("00:00:00:00:00:01");
    packet->PeekHeader(h );
    return h.GetAddr1()==addr; 
}


WifiMode YansWifiPhyAp::getRate()
{
    return currentRate;
}

}//ns namespace
