/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005,2006 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Nikos Stathopoulos<stathopou@ceid.upatras.gr>
 */

#include "yans-wifi-phy-sta.h"
#include "yans-wifi-channel.h"
#include "wifi-mode.h"
#include "wifi-preamble.h"
#include "wifi-phy-state-helper.h"
#include "error-rate-model.h"
#include "ns3/simulator.h"
#include "ns3/packet.h"
#include "ns3/assert.h"
#include "ns3/log.h"
#include "ns3/double.h"
#include "ns3/uinteger.h"
#include "ns3/enum.h"
#include "ns3/pointer.h"
#include "ns3/net-device.h"
#include "wifi-mac.h"
#include "ap-wifi-mac.h"
#include "sta-wifi-mac.h"
#include "adhoc-wifi-mac.h"
#include "wifi-net-device.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/boolean.h"
#include "snr-tag.h"
#include "signal-tag.h"
#include "rate-tag.h"
#include "noise-tag.h"
#include "yans-wifi-phy.h"
#include <cmath>
#include "ns3/ipv4-header.h"
#include "ns3/core-module.h"
#include<string.h>

#define HIGH 1
#define LOW 0
#define PACKET_LOST 2
#define VARIATION_THRES 1

#define PACKET_COUNT 10
#define TOLERANCE 2
#define TX_MAX 20
#define TX_MIN 5

NS_LOG_COMPONENT_DEFINE ("YansWifiPhySta");

namespace ns3 {

#define HT_PHY 127 
    
NS_OBJECT_ENSURE_REGISTERED (YansWifiPhySta);

bool fromAp(const Ptr<const Packet> &packet);

TypeId
YansWifiPhySta::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::YansWifiPhySta")
    .SetParent<YansWifiPhy> ()
    .AddConstructor<YansWifiPhySta> ()
    .AddTraceSource("MechanismEvent",
                    "Trace Source used for the mechanism",
                     MakeTraceSourceAccessor(&YansWifiPhySta::m_phyMechanismTrace)) 
    .AddTraceSource("RateMechanismEvent",
                    "Trace Source used for the rate mechanism",
                     MakeTraceSourceAccessor(&YansWifiPhySta::m_phyRateMechanismTrace)) 
  ;
  return tid;
}

YansWifiPhySta::YansWifiPhySta () :YansWifiPhy()
{
  NS_LOG_FUNCTION (this);
NS_LOG_DEBUG("class STA");
    SetTxPowerStart(20);
    SetTxPowerEnd(20);  
    m_power=20;
    m_enable=true;
    sendVariation=false;
    m_packetCounter = 0;  
    perf_snr=24.0;
    m_snrSta=0;
    m_signalSta=0;
    m_snrAp =0;
    m_lastSnrAp=0;
    m_drop=0;
    m_received_corr=0;
    m_sent=0;
    
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate6_5MbpsBW20MHz(), 5.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate13MbpsBW20MHz(), 8.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate19_5MbpsBW20MHz(), 10.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate26MbpsBW20MHz(), 14.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate39MbpsBW20MHz(), 17.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate52MbpsBW20MHz(), 22.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate58_5MbpsBW20MHz(), 23.0));
    rates.push_back(MechanismRate(WifiPhy::GetOfdmRate65MbpsBW20MHz(), 24.0));
    NS_LOG_DEBUG("class sta");
}

YansWifiPhySta::~YansWifiPhySta ()
{
 
    NS_LOG_DEBUG("STA RECEIVED:"<<m_received_corr<<" STA SENT:"<<m_sent);
  NS_LOG_FUNCTION (this);
}

InterferenceHelper::SnrPer
YansWifiPhySta::GetSnrPer (Ptr<InterferenceHelper::Event> event)
{
    return m_interference.CalculateSnrPer(event);
}

void
YansWifiPhySta::SendPacket (Ptr<const Packet> packet, WifiTxVector txVector, enum WifiPreamble preamble)
{
    if(sendVariation)
    {
        int r=calcRate();
        SignalTag snrVariationTag;
        snrVariationTag.Set((double) r);
        if(! packet->PeekPacketTag(snrVariationTag))
             packet->AddByteTag(snrVariationTag);
        sendVariation=false;
    }
    m_sent++;
    YansWifiPhy::SendPacket(packet,txVector,preamble);
}

double YansWifiPhySta::GetSnrSta() const
{
    return m_snrSta;
}

double YansWifiPhySta::GetSnrAp() const
{
    return m_snrAp;
}


double YansWifiPhySta::GetTxPower() const
{
    return m_power;
}

double YansWifiPhySta::GetVariation() const
{
    return m_snrVariation;
}

void
YansWifiPhySta::NotifyMechanism(Ptr<const Packet> packet, Ptr<InterferenceHelper::Event> event){
      m_phyMechanismTrace(packet, this, event);
}

void
YansWifiPhySta::NotifyRateMechanism(Ptr<const Packet> packet, Ptr<InterferenceHelper::Event> event, int action){
      m_phyRateMechanismTrace(packet, this, event, action);
}

void
YansWifiPhySta::CalculateSnr(){
    
    m_lowSnrAp=1000;
    m_snrAp = 0;
    
    for(int i =0 ; i <(int) m_snrApList.size(); i++)
    {
        m_snrAp +=m_snrApList.at(i);
        if(m_lowSnrAp>m_snrApList.at(i) )
            m_lowSnrAp=m_snrApList.at(i);
    }
    m_snrAp/=m_snrApList.size();

}

void YansWifiPhySta::CalculateVariation(){
    m_snrVariation = 0;
    
    for(int i = 0; i < (int)m_snrApList.size() && i<5; i++ )
    {
        m_snrVariation += pow(m_snrApList.at(i) - m_snrAp, 2);
    }
    m_snrVariation = sqrt(m_snrVariation/(int)m_snrApList.size());
}


void
YansWifiPhySta::EndReceive (Ptr<Packet> packet, Ptr<InterferenceHelper::Event> event)
{

   
  NS_LOG_FUNCTION (this << packet << event);
  NS_ASSERT (IsStateRx ());
  NS_ASSERT (event->GetEndTime () == Simulator::Now ());
  struct InterferenceHelper::SnrPer snrPer = m_interference.CalculateSnrPer (event);
  m_interference.NotifyRxEnd ();
  double signalDbm=0;
  double failure=false;
//   if(event->GetPayloadMode ().GetDataRate () == 65000000)
//   NS_LOG_DEBUG ("mode=" << (event->GetPayloadMode ().GetDataRate ()) <<
//                 ", snr=" << 10.0 * std::log10(snrPer.snr) << ", per=" << snrPer.per << ", size=" << packet->GetSize ());
  bool packetFromAp=fromAp(packet);

  if (m_random->GetValue () > snrPer.per)
    {
      NotifyRxEnd (packet);
      uint32_t dataRate500KbpsUnits = event->GetPayloadMode ().GetDataRate () * event->GetTxVector().GetNss()/ 500000;
      bool isShortPreamble = (WIFI_PREAMBLE_SHORT == event->GetPreambleType ());
      signalDbm = RatioToDb (event->GetRxPowerW ()) + 30;
      
      double noiseDbm = RatioToDb (event->GetRxPowerW () / snrPer.snr) - GetRxNoiseFigure () + 30;
      NotifyMonitorSniffRx (packet, (uint16_t)GetChannelFrequencyMhz (), GetChannelNumber (), dataRate500KbpsUnits, isShortPreamble, signalDbm, noiseDbm);
      m_state->SwitchFromRxEndOk (packet, snrPer.snr, event->GetPayloadMode (), event->GetPreambleType ());
      
    }
  else //failure
  {
//       NS_LOG_DEBUG("FAIL");
      NotifyRxDrop (packet);
      m_state->SwitchFromRxEndError (packet, snrPer.snr);  
      failure=true;      
   }
        
    if(packetFromAp)
    {    
        m_packetCounter++;
        m_lastSnrAp=RatioToDb(snrPer.snr);
        AddSnrValue(m_lastSnrAp);
        if(m_packetCounter%PACKET_COUNT==0)
        {
            CalculateSnr();
            CalculateVariation();
            m_snrApList.clear();
            if(GetVariation()>1 )
                sendVariation=true;
        }    
        
        if(!failure)
            TransmissionControl(packet,event);
        
        if( failure )
        {
            NotifyRateMechanism(packet,event, PACKET_LOST);
            m_drop++;
        }
        else
        {
            m_received_corr++;
            NotifyMechanism(packet,event);
        }
    }    
}

void YansWifiPhySta::TransmissionControl(Ptr< Packet >& packet, Ptr< InterferenceHelper::Event >& event)
{
    
    SnrTag snrTag;
    SignalTag signalTag;
    if(!packet->FindFirstMatchingByteTag(snrTag) )
        return ;
    
    if(!packet->FindFirstMatchingByteTag(signalTag) )
        return ;    
    
    m_snrSta = snrTag.Get(); 
    
    if(m_enable)
    {
        if( m_snrSta < perf_snr - TOLERANCE || m_snrSta >perf_snr + TOLERANCE) 
        {        
            double signalDbm = signalTag.Get();
            double pathLoss = m_power * GetTxGain() - signalDbm;
            double noiseDbm = signalDbm-m_snrSta;
        
            int  newTx = ceil(perf_snr + noiseDbm + pathLoss);        
            SetPower(newTx);   
        }
    }
    
}

double YansWifiPhySta::SetPower(double newTx) 
{        
    
    if(newTx > TX_MAX)
        newTx = TX_MAX;
    else if(newTx < TX_MIN)
        newTx = TX_MIN;    
    
    SetTxPowerEnd(newTx);
    SetTxPowerStart(newTx);
    m_power=newTx;
    
    return newTx;
}

void YansWifiPhySta::AddSnrValue(double snr)
{    
    std::vector<double>::iterator it = m_snrApList.begin();
    m_snrApList.insert(it, snr);
}

double YansWifiPhySta::GetLastSnrAp() const
{
    return m_lastSnrAp;
}

int YansWifiPhySta::calcRate()
{
    double min=1000;//big number
    int j=-1;
    for (uint32_t i = 0; i < rates.size(); i++) 
    {
        double tmp=rates.at(i).GetSnr() - m_lowSnrAp;
        if(tmp<0)
            tmp*=-1;        
        
        if(tmp<min)
        {
            min=tmp;
            j=i;
        }
    }   
    
    if(j!=0)
        j--;
        
    return j;
}


bool fromAp(const Ptr<const Packet> &packet) 
{
    WifiMacHeader h;
    Mac48Address addr("00:00:00:00:00:02");
    packet->PeekHeader(h);
    return h.GetAddr1()==addr;
}

} // namespace ns3

